use super::main_axis_alignment::MainAxisAlignment;

pub struct Spacing {
	alignment: MainAxisAlignment,
	extra: usize,
	n_children: usize,
	index: usize,
	equal_space: usize,
	remainder: usize,
}

impl Spacing {
	/// Given the provided extra space and children count,
	/// this returns an iterator of `f64` spacing,
	/// where the first element is the spacing before any children
	/// and all subsequent elements are the spacing after children.
	pub(super) fn new(alignment: MainAxisAlignment, extra: usize, n_children: usize) -> Spacing {
		let equal_space = if n_children > 0 {
			match alignment {
				MainAxisAlignment::Center => extra / 2,
				MainAxisAlignment::SpaceBetween => extra / (n_children - 1).max(1),
				MainAxisAlignment::SpaceEvenly => extra / (n_children + 1),
				MainAxisAlignment::SpaceAround => extra / (2 * n_children),
				_ => 0,
			}
		} else {
			0
		};
		Spacing {
			alignment,
			extra,
			n_children,
			index: 0,
			equal_space,
			remainder: 0,
		}
	}

	fn next_space(&mut self) -> usize {
		let desired_space = self.equal_space + self.remainder;
		let actual_space = desired_space;
		self.remainder = desired_space - actual_space;
		actual_space
	}
}

impl Iterator for Spacing {
	type Item = usize;

	fn next(&mut self) -> Option<usize> {
		if self.index > self.n_children {
			return None;
		}
		let result = {
			if self.n_children == 0 {
				self.extra
			} else {
				#[allow(clippy::match_bool)]
				match self.alignment {
					MainAxisAlignment::Start => match self.index == self.n_children {
						true => self.extra,
						false => 0,
					},
					MainAxisAlignment::End => match self.index == 0 {
						true => self.extra,
						false => 0,
					},
					MainAxisAlignment::Center => match self.index {
						0 => self.next_space(),
						i if i == self.n_children => self.next_space(),
						_ => 0,
					},
					MainAxisAlignment::SpaceBetween => match self.index {
						0 => 0,
						i if i != self.n_children => self.next_space(),
						_ => match self.n_children {
							1 => self.next_space(),
							_ => 0,
						},
					},
					MainAxisAlignment::SpaceEvenly => self.next_space(),
					MainAxisAlignment::SpaceAround => {
						if self.index == 0 || self.index == self.n_children {
							self.next_space()
						} else {
							self.next_space() + self.next_space()
						}
					}
				}
			}
		};
		self.index += 1;
		Some(result)
	}
}
